import java.awt.*;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.List;
import java.util.Queue;
import java.util.*;

public class Maze {

    // Direction vectors
    // IN ORDER: UP (-1,0), RIGHT (0,1), DOWN 1,0), LEFT (0, -1)
    public static final int[] ROW_OFFSET = {-1, 0, 1, 0};
    public static final int[] COL_OFFSET = { 0, 1, 0, -1 };

    // Node for a maze corridor
    static class MazeNode {
        List<MazeNode> adjacencyList;
        Point coords;

        public MazeNode() {
            adjacencyList = new ArrayList<>();
        }

        // Adds an undirected path between this node and another node
        public void addNeighbor(MazeNode node) {
            this.adjacencyList.add(node);
            node.adjacencyList.add(this);
        }
    }

    private final String[][] mazeArray;
    private final HashMap<Point, MazeNode> nodes;

    public Maze(String mazeFile) {
        nodes = new HashMap<>();
        BufferedReader br = null;
        try {
            br = new BufferedReader(new FileReader(mazeFile));
            String[] header = br.readLine().split(" ");
            this.mazeArray = new String[Integer.parseInt(header[0])][Integer.parseInt(header[1])];
            buildMazeArray(br);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Builds a maze array from a file
     */
    private void buildMazeArray(BufferedReader br) {
        String line;
        int row = 0;
        try {
            while ((line = br.readLine()) != null) {
                String[] lineItems = line.split(" ");
                System.arraycopy(lineItems, 0, mazeArray[row], 0, lineItems.length);
                row++;
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Gets the map of nodes
     * @return hashmap of points to their respective nodes
     */

    public HashMap<Point, MazeNode> getNodes() {
        return nodes;
    }

    /**
     * TODO: Create an undirected graph from the maze array.
     * A maze array has been created for you in the constructor.
     * You must take this array and turn it into an undirected graph
     * containing only the corridors (".").
     * Note: the Point class (imported for you) is used at the key
     * to the hashmap called 'nodes'. You may pass in a new instance of
     * Point with the same coordinates, and it will be interpreted as the
     * same coordinates. For instance,
     *      nodes.put(new Point(1,1), new MazeNode)
     *      Point coords = new Point(1, 1);
     *      nodes.get(coords)
     * Is a valid way to get the original MazeNode.
     */
    public void createMazeGraph() {
        // TODO
    }

    /**
     * BFS method to find the nearest exit to the maze
     * @param startingCoords starting coordinates
     * @return the nearest node that represents the exit of the maze
     */
    public MazeNode findNearestExit(Point startingCoords) {
        Queue<Point> q = new LinkedList<>();
        q.add(startingCoords);

        List<Point> visited = new ArrayList<>();
        visited.add(startingCoords);

        while (!q.isEmpty()) {
            Point currCoord = q.poll();
            int numAdj = 0;
            for (MazeNode node: nodes.get(currCoord).adjacencyList) {
                Point adjCoord = node.coords;

                if (!visited.contains(adjCoord)) {
                    q.add(adjCoord);
                    visited.add(adjCoord);
                    numAdj++;
                }
            }
            // Exit found
            if (numAdj == 0) {
                return nodes.get(currCoord);
            }
        }
        return null;
    }
}
